from __future__ import print_function
import os
import sys
import logging
import argparse
from bam2x import TableIO,IO,DBI
from bam2x.Tools import seq_wrapper,compatible,overlap,_translate_to_meta,translate,compatible_with_transcript
def help():
    return "query RNA Seq bam file , and report it in Transcript coordinates. \n input file is a Bed12 format."
def set_parser(parser):
    parser.add_argument("-b","--bam",type=str,dest="bam",help="bam file")
    parser.add_argument("-s","--strand",type=str,dest="strand",choices=["read1","read2"],default="read2",help="strand of RNA SEQ library [read1 or read2] , default %(default)s")
    parser.add_argument("-m",dest="hit",action="store_true",default=False,help="mininum report, only report comptible reads within transcript")
    
def run(args):
    #logging.basicConfig(level=logging.DEBUG)
    dbi=DBI.init(args.bam,"bam")
    out=IO.fopen(args.output,"w")
    for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
        print ("QR\t",i,file=out)
        for j in dbi.query(i,method="bam1",strand=args.strand):
            if compatible_with_transcript(j,i):
                print ("HT\t",_translate_to_meta(i,j),file=out)
            elif not args.hit:
                print ("OP\t",j,file=out)
        print ("",file=out)

if __name__=="__main__":
    from bam2x.IO import parser_factory
    p=parser_factory(description=help())
    set_parser(p)
    if len(sys.argv)==1:
        print(p.print_help())
        exit(0)
    run(p.parse_args())

 



# Programmer : zhuxp
# Date: 
# Last-modified: 02-12-2014, 12:43:09 EST
__all__=['BamI','TabixI','MetaDBI',"BinIndexI","BamlistI","TwoBitI","GenomeI"]
from DB import *

FormatToDBI = { 
             "flatfile":BinIndexI,
             "binindex":BinIndexI,
             "bam":BamI,
             "bams":BamlistI,
             "bamlist":BamlistI,
             "tabix":TabixI,
             "genome":GenomeI,
             "bigwig":BigWigI
            }
def query(x,dbi,**dict):
    '''
    DBI.query is wrapper for query all kinds of data|file|database. 
    '''
    for i in dbi.query(x,**dict):
        yield i
        
def init(handle,dbformat,**dict):
    '''
    A Wrapper for initialize the DBI
    '''
    if dbformat in FormatToDBI:
        dbi=FormatToDBI[dbformat]
        return dbi(handle,**dict)
    else:
        dbi=BinIndexI
        return dbi(handle,**dict)




from __future__ import print_function
import os
import sys
import logging
import argparse
from bam2x import TableIO,IO,DBI
from bam2x.Tools import seq_wrapper,compatible,overlap,_translate_to_meta,translate,compatible_with_transcript

def help():
    return "query RNA Seq bam file , and report RPKM ( or unique RPKM ) only compatible reads are counted. \n input file is a Bed12 format."
def set_parser(parser):
    parser.add_argument("-b","--bam",type=str,dest="bam",help="bam file")
    parser.add_argument("-s","--strand",type=str,dest="strand",choices=["read1","read2"],default="read2",help="strand of RNA SEQ library [read1 or read2] , default %(default)s")
    parser.add_argument("-u",dest="uniq",action="store_true",default=False,help="only report unique mapped reads")
    
def run(args):
    #logging.basicConfig(level=logging.DEBUG)
    dbi=DBI.init(args.bam,"bam")
    mapped=dbi.mapped
    out=IO.fopen(args.output,"w")
    for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
        print(i.id,"\t",file=out)
        s=0.0
        l=i.cdna_length()
        if args.uniq:
            for j in dbi.query(i,method="bam1",strand=args.strand,uniq=args.uniq):
                if compatible_with_transcript(j,i):
                    s+=1.0
        else:
            for j in dbi.query(i,method="bam1",strand=args.strand,uniq=args.uniq):
                if compatible_with_transcript(j,i):
                    (nh,_,_)=j.itemRgb.split(",")
                    nh=int(nh)
                    s+=1.0/nh
        rpkm=float(s)*(1000000.0/mapped)*(1000.0/float(l))
        print(rpkm,file=out)

if __name__=="__main__":
    from bam2x.IO import parser_factory
    p=parser_factory(description=help())
    set_parser(p)
    if len(sys.argv)==1:
        print(p.print_help())
        exit(0)
    run(p.parse_args())

 



__all__=["getseq","query_bam","read","sort","cmpgene","getanno","query_RNASeq","load_bed_to_db","query_db","query_RNA_RPKM","tabix"]
from . import *

'''
command lines
'''
commands= {
    "getseq":getseq,
    "getanno":getanno,
    "query_bam":query_bam,
    "query_RNASeq":query_RNASeq,
    "query_RNA_RPKM":query_RNA_RPKM,
    "read":read,
    "sort":sort,
    "cmpgene":cmpgene,
    "load_bed_to_db":load_bed_to_db,
    "query_db":query_db,
    "tabix":tabix
}





#!/usr/bin/env python
from __future__ import print_function
# Programmer : zhuxp
# Date: 
# Last-modified: 03-17-2014, 18:05:04 EDT
import os,sys,argparse
from bam2x import TableIO,Tools
from bam2x import IO
from bam2x.Annotation import BED12 as Bed12
import sqlite3
import string
from bam2x.DBI.Templates import select_template as template
from bam2x.DBI.Templates import factories
#template=string.Template("""select * from $table_name where name=\"$name\"""")

def set_parser(p):
    ''' This Function Parse the Argument '''
    p.add_argument("-d","--database",dest="db",type=str,help="database file name",default="guess")
    p.add_argument("-D","--database_format",dest="db_format",type=str,choices=factories.keys(),help="database entry format",default="bed12")
    p.add_argument("-t","--table_name",dest="table_name",type=str,help="table name",default="test")
def help():
    return "query sqlite3 database to get a transcipt, and then do further analysis. this is a demo program."
#def factory(cursor,r):
#    return Bed12._make(Bed12._types(r[1:]))
def run(args):
    db_filename=args.db
    out=IO.fopen(args.output,"w")
    if os.path.exists(args.input):
        fin=IO.fopen(args.input,"r")
    else:
        fin=(args.input,)
    if not os.path.exists(db_filename):
        print("can't find database %s"%db_filename,file=sys.stderr)
        exit(1)
    print("# Database file : %s"%db_filename,file=out)
    with sqlite3.connect(db_filename) as conn:
        conn.row_factory=factories[args.db_format]
        cursor=conn.cursor()
        for i in fin:
            i=i.strip()
            i=i.strip(" ")
            print("# query %s"%i,file=out)
            s=template.substitute({"table_name":args.table_name,"name":i})
            print("# "+s,file=out)
            cursor.execute(s)
            r=None
            try:
                r=cursor.fetchone()
                print(r,file=out)
            except:
                raise
        
if __name__=="__main__":
    from bam2x.IO import parser_factory
    p=parser_factory(description=help())
    set_parser(p)
    if len(sys.argv)==1:
        print(p.print_help())
        exit(0)
    run(p.parse_args())



from __future__ import print_function
import os
import sys
import logging
import argparse
from bam2x import TableIO,IO,DBI
from bam2x.Tools import seq_wrapper
def help():
    return "get [utr, cds, exon,intron,upstream or downstream] annotations from bed12 file"
def set_parser(parser):
    parser.add_argument("-a","--anno",type=str,choices=("cds","utr5","utr3","exon","intron","utr","upstream","downstream"),dest="annotation",default="cdna")
    parser.add_argument("--bp",type=int,dest="bp",default=1000,help="upstream or downstream bp number")
    
def run(args):
    out=IO.fopen(args.output,"w")
    if args.annotation=="exon":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            for j in i.Exons():
                print(j,file=out)
    elif args.annotation=="intron":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            for j in i.Introns():
                print(j,file=out)
    elif args.annotation=="cds":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            j=i.cds()
            if j is not None and j.cdna_length() > 0:
                print(j,file=out)

    elif args.annotation=="cds":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            j=i.cds()
            if j is not None and j.cdna_length() > 0:
                print(j,file=out)

    elif args.annotation=="utr5":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            j=i.utr5()
            if j is not None and j.cdna_length() > 0:
                print(j,file=out)

    elif args.annotation=="utr3":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            j=i.utr3()
            if j is not None and j.cdna_length() > 0:
                print(j,file=out)
    elif args.annotation=="utr":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            j=i.utr5()
            if j is not None and j.cdna_length() > 0:
                print(j,file=out)
            j=i.utr3()
            if j is not None and j.cdna_length() > 0:
                print(j,file=out)
    elif args.annotation=="upstream":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            j=i.upstream(args.bp)
            print(j,file=out)
    elif args.annotation=="downstream":
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed12"):
            j=i.downstream(args.bp)
            print(j,file=out)



if __name__=="__main__":
    from bam2x.IO import parser_factory
    p=parser_factory(description=help())
    set_parser(p)
    if len(sys.argv)==1:
        print(p.print_help())
        exit(0)
    run(p.parse_args())




    











# Programmer : zhuxp
# Date: 
# Last-modified: 02-12-2014, 22:02:57 EST
import os,sys,argparse,types
from bam2x.Annotation import BED6 as Bed 
from bam2x import TableIO
'''
binindex Data Structure

self.data is hash table. 
Key is Chromosome
self.data["chr1"] is a list of feature list
self.data["chr"][bin_number] is a list of features that have bin index.

The Most Useful Function is:

Read features into a data structure:
    Example:
        f=TableIO.parse("filename","bed12")
        data=binindex(f)
    Example:
        f=TableIO.parse("filename","vcf")
        data=binindex(f)
    Example:
        data=binindex(bedlist)
    

Query the overlap features in the data structure:
    Usage:
        for i in data.query(bed):
            print i

This BinIndex Data Structure was wrapped into DBI.init() and DBI.query():
    Usage Example:
        dbi=DBI.init("file.bed","bed")
        x=Bed(chr="chr1",start=1,stop=200)
        for i in dbi.query(x):
            print i
TO DO:
    Now BinIndex only have item inherent from Bed Class, which means it need to have at least three attributes, chr, start and stop. and start is 0-index.  
    it use the fourth attribute, id.
    Need to improve to read any table that contain this information but not not in a pre defined format. 
'''
class binindex(object):
	'''
        binindex data structure
        this data structure hold genome annotations. 
        it can be queried overlap features with given chromosome regions.
        
        Examples:
        
        Query Overlap Features:
        
        data=binindex(TableIO.parse("file.bed","bed"))
        bed=Bed(chr="chr1",start=1,stop=10000)
        for i in data.query(bed):
            print i
        Add Two Data Struture:
            dataC = dataA + dataB
        
        '''
	binOffsets=(512+64+8+1,64+8+1,8+1,1,0)
	binFirstShift=17
	binNextShift=3
        binLength=(4096+512+64+8+1)
        @staticmethod
	def range2bin(start,end):
	    '''
	        Convert (start,end) to bin index
	        
	        Usage:
	           data=[[] for row in range(4096+512+64+8+1)]
	           for i in list:
	               bin=range2bin(i.start,i.stop)
	               data[bin].append(i)
	    '''
	    startBin=start
	    endBin=end-1 
	    startBin >>= binindex.binFirstShift
	    endBin >>= binindex.binFirstShift
	    for i in range(len(binindex.binOffsets)):
	        if startBin==endBin:
	            return binindex.binOffsets[i]+startBin
	        startBin >>= binindex.binNextShift
	        endBin >>= binindex.binNextShift
	    print >>sys.stderr,"start %d , end %d out of range"%(start,end) 
	    return None
            
        @staticmethod
	def iter_range_overlap_bins(start,end):
	    '''
	        Iterate the possible overlap binindex
	        the core function of query overlap features.
	        Usage:
	            for i in binindex.iter_range_overlap_bins(start,end):
	                for j in data[i]:
	                    if start > j.stop and stop < j.start:
	                        print j
	    '''
	    startBin=start
	    endBin=end-1
	    startBin >>= binindex.binFirstShift
	    endBin >>= binindex.binFirstShift
	    for i in range(len(binindex.binOffsets)):
	        for j in range(startBin,endBin+1):
	            yield j+binindex.binOffsets[i]
	        startBin >>= binindex.binNextShift
	        endBin >>= binindex.binNextShift
        @staticmethod
        def bin2cov(data):
            '''
            input a binindex data
            output total length of coverage for each bin 
            '''
            coverage={}
            for index,i in enumerate(data):
                if index%100==0:
                    print >>sys.stderr,"processing %d entries\r"%index,
                if not coverage.has_key(i.chr):
                    coverage[i.chr]=[0 for k in range(binindex.binLength)]
                i_bin=binindex.range2bin(i.start,i.stop)
                for j in binindex.iter_range_overlap_bins(i.start,i.stop):
                    (start,stop)=binindex.bin2range(j)
                    max_start=max(start,i.start)
                    min_stop=min(stop,i.stop)
                    length=min_stop-max_start
                    if float(length)/(stop-start)<0:
                        print >>sys.stderr,length,stop,start
                    coverage[i.chr][j]+=length
            return coverage
        
        
        @staticmethod
        def bin2range(bin):
            binShift=binindex.binFirstShift
            for i in binindex.binOffsets:
                if bin-i>=0:
                    bin=bin-i
                    break
                binShift+=binindex.binNextShift
            return (bin<<binShift,(bin+1)<<binShift)
        
        @staticmethod
        def bin2length(bin):
            (start,end)=binindex.bin2range(bin)
            return end-start
        @staticmethod
        def bin2level(bin):
            for (i,x) in enumerate(binindex.binOffsets):
                if bin-x>=0:
                    return 4-i 

      
        
        def __init__(self,handle=None,**kwargs):
            self.data={}
            if handle is not None:
                self.read(handle,**kwargs)
            
	def append(self,bed):
	    '''
	    data is a binindex Data Structure
	    bed is an anntation feature. ( bed vcf or genebed etc.)
	    append an annotation into data
	    '''
	    a=bed
	    bin=binindex.range2bin(a.start,a.stop)
	    if not self.data.has_key(a.chr):
	        self.data[a.chr]=[[] for row in range(binindex.binLength)]
	    self.data[a.chr][bin].append(a)
	
	def remove(self,bed):
	    '''
	    delete bed entry from a BinIndex Data Structure data.
	    '''
	    a=bed
	    bin=binindex.range2bin(a.start,a.stop)
	    for i,x in enumerate(self.data[a.chr][bin]):
	        if a==x:                          # use define equal in Bed.__cmp__
	            del self.data[a.chr][bin][i]
	
	def merge(self,other):
	    '''
	    merge two binindex data structure
	    and only ratain one if duplicates.
	    '''
	    for i in other:
	        flag=1
	        for j in self.query(i):
	            if j.start==i.start and j.stop==i.stop and j.id==i.id:
	                flag=0
	        if flag:
	            self.append(i)
        def __add__(self,b):
            '''
            overload + 
            merge two data structure
            and retain all information including duplicate entries.
            '''
            c=binindex()
            for i in self:
                c.append(i)
            for i in b:
                c.append(i)
            return c

	
	    
	def __iter__(self):
	    '''
	    iter all the features in binindex structure
	    Usase:
	        data is a binindex data structure
	
	        for i in iterBinIndex(data):
	            print i
            should revise to iter in sorted order
	    '''
	    chrs=self.data.keys()
	    chrs.sort()
	    for chr in chrs:
	        for binindex in range(4096+512+64+8+1):
	            for i in self.data[chr][binindex]:
	                yield i
	
	def read(self,handle,**kwargs):
	    '''
	    Read iterator into BinIndex Data Structure
	    if have cls, turn tuple into namedtuple
            '''
            cls=None
            if kwargs.has_key("cls"):
                cls=kwargs["cls"]
	    if cls is not None:
                for i in handle:      
                    a=cls._make(cls._types(i))
	            self.append(a)
            else:
                for i in handle:
                    self.append(i)
	
	def query(self,bed,**kwargs):
	    '''
	    iterator the bed overlap features in data
	    Usage:
                data=bedindex(file.bed)
	        for i in data.query(bed):
	            print i
	
	    **kwargs for further extension for annotation with no pre define format
	    '''
	
	    if type(bed)==type((1,2,3)) or type(bed)==([1,2,3]):
	        bed=Bed(*bed[0:3])  # guess (chrome,start,stop)
	    if not self.data.has_key(bed.chr):
	         raise StopIteration 
	    D=self.data[bed.chr]
	    for bin in binindex.iter_range_overlap_bins(bed.start,bed.stop):
	        for f in D[bin]:
	            if f.start < bed.stop and f.stop > bed.start:
	                yield f
	    raise StopIteration
        def __len__(self):
            s=0
            for chr in self.data:
                for j in self.data[chr]:
                    s+=len(j)
            return s
        def __str__(self):
            s="binindex data structure object with "+str(len(self))+" entries"
            return s
        def uniq(self):
            '''
            return a binindex data structure that have no duplicate entries.

            Usage:
            dataUniq=data.uniq()
            '''
            c=binindex()
            for i in self:
                if c.contains(i): continue
                else:
                    c.append(i)
            return c
        def contains(self,a):
            '''
            binary decision that if this data struct has feature a.
            '''
            if not self.data.has_key(a.chr):
                return False
	    bin=binindex.range2bin(a.start,a.stop)
	    for i,x in enumerate(self.data[a.chr][bin]):
	        if a==x:
                    return True
            return False


	
	
	
def test():
    if len(sys.argv)==1:
        print >>sys.stderr,"Usage: binindex.py file.bed"
        exit()
    a=TableIO.parse(sys.argv[1],'bed12')
    data=binindex(a)
    data2=binindex()
    bed=Bed("chr1",100000,2000000,".",".",".")
    for i in data.query(bed):
        print "before remove:",len(data)
        data.remove(i)
        print "after remove:",len(data)
        data2.append(i)
        print data2
    for i in data2:
        print i
    print "data finalize:"
    data.merge(data2)
    print data
    print data+data2
    print data
    print data.uniq()
    print data

    
    
if __name__=="__main__":
    test()





from __future__ import print_function
import os
import sys
import logging
import argparse
import sqlite3
from os.path import splitext
import string
from bam2x.Tools import reverse_translate
from bam2x.DBI.Templates import bed12_schema_template as schema_t
from bam2x.DBI.Templates import bed12_insert_template as insert_t
import logging
from bam2x import TableIO,IO
from bam2x.Annotation import BED12 as Bed12
def help():
    return "translate bed from gene coordinates to chromosome coordinates"
def set_parser(parser):
    parser.add_argument("-t","--translator",type=str,dest="translator",help="translator bed file or db[sqlite3] (gene coordinates in chromosome)")
    parser.add_argument("-T","--table_name",type=str,dest="table_name",default="test")
    
def _generate_db(filename,db_filename,table_name):
    with sqlite3.connect(db_filename) as conn:
        cursor=conn.cursor()
        S=schema_t.substitute({"table_name":table_name})
        cursor.execute(S)
        LOAD_S=insert_t.substitute({"table_name":table_name})
        s=TableIO.parse(IO.fopen(filename,"r"),"simple")
        cursor.executemany(LOAD_S,s)
        conn.commit()

template=string.Template("""select * from $table_name where name='$name'""")
def run(args):
    logging.basicConfig(level=logging.DEBUG)
    db_filename=args.translator
    t_name,t_ext=splitext(args.translator)
    '''
    test if it is db file
    generate db file if it doesn't exists.
    '''
    if t_ext!="db":
        #possible_db=args.translator.strip("\\.gz")+".db"
        possible_db=args.translator+".db"
        print(possible_db)
        if os.path.exists(possible_db):
            db_filename=possible_db
        else:
            _generate_db(args.translator,possible_db,args.table_name)
            db_filename=possible_db
    
    '''
    query db file
    '''
    out=IO.fopen(args.output,"w")
    with sqlite3.connect(db_filename) as conn:
        conn.row_factory=lambda conn,x: Bed12._make(Bed12._types(x[1:]))
        cursor=conn.cursor()
        for i in TableIO.parse(IO.fopen(args.input,"r"),"bed"):
            s=template.substitute({"table_name":args.table_name,"name":i.chr.strip()})
            print(s)
            cursor.execute(s)
            gene=None
            try:
                gene=cursor.fetchone()
                logging.debug(i)
                logging.debug(i.cdna_length())
                logging.debug(gene)
                logging.debug(gene.cdna_length())
            except:
                raise
                logging.warning("can't find gene %s"%i.chr)
                continue
            assert gene.cdna_length() > i.cdna_length()
            print(reverse_translate(gene,i),file=out) 
    

if __name__=="__main__":
    from bam2x.IO import parser_factory
    p=parser_factory(description=help())
    set_parser(p)
    if len(sys.argv)==1:
        print(p.print_help())
        exit(0)
    run(p.parse_args())










#!/usr/bin/env python
from __future__ import print_function
# Programmer : zhuxp
# Date: 
# Last-modified: 03-17-2014, 15:59:20 EDT
import os,sys,argparse
from bam2x import TableIO,Tools
from bam2x import IO
import sqlite3
import string
from bam2x.DBI.Templates import schema_templates
from bam2x.DBI.Templates import insert_templates

def set_parser(p):
    ''' This Function Parse the Argument '''
    p.add_argument("-D","--database",dest="db",type=str,help="database file name",default="guess")
    p.add_argument("-I","--input_format",dest="input_format",choices=schema_templates.keys(),type=str,help="input format , default: %(default)s",default="bed12")
    p.add_argument("-t","--table_name",dest="table_name",type=str,help="table name, default: %(default)s",default="test")
def help():
    return "read bed12 file into a sqlite3 db. if sqlite3 db is not exists , it will generate one"
    
def run(args):
    schema_template=schema_templates[args.input_format]
    SQL_template=insert_templates[args.input_format]
    db_filename=args.db
    out=IO.fopen(args.output,"w")
    if db_filename=="guess":
        db_filename=args.input.strip(".gz")+".db"
    db_is_new = not os.path.exists(db_filename)
    print("Database file : %s"%db_filename,file=out)
    with sqlite3.connect(db_filename) as conn:
        cursor=conn.cursor()
        if db_is_new:
            print ('Creating table %s if not exists\n________________________________'%args.table_name,file=out)
            S=schema_template.substitute({"table_name":args.table_name})
            print (S,file=out)
            print ("_______________________________",file=out)
            cursor.execute(S)
        fin=IO.fopen(args.input,"r")
        S1=SQL_template.substitute({"table_name":args.table_name})
        print(S1,file=out)
        s=TableIO.parse(args.input,"simple")
        cursor.executemany(S1,s)
        conn.commit()
        print("loaded",file=out)

        
if __name__=="__main__":
    from bam2x.IO import parser_factory
    p=parser_factory(description=help())
    set_parser(p)
    if len(sys.argv)==1:
        print(p.print_help())
        exit(0)
    run(p.parse_args())



from __future__ import print_function
import os
import sys
import logging
import argparse
from bam2x import TableIO,IO,DBI
from bam2x.Tools import seq_wrapper
def help():
    return "get sequence from twobit genome file"
def set_parser(parser):
    parser.add_argument("-m",type=str,choices=("seq","cDNA","cdna","cds","utr5","utr3"),dest="method",default="seq")
    parser.add_argument("-g","--genome",type=str,dest="genome",help="chromsome.2bit file")
    parser.add_argument("-b","--bed_column_number",type=int,choices=[3,6,12],dest="bed_column_number",default=12) 
    
def run(args):
    bedformat="bed"+str(args.bed_column_number)
    dbi=DBI.init(args.genome,"genome")
    out=IO.fopen(args.output,"w")
    for i in TableIO.parse(IO.fopen(args.input,"r"),bedformat):
        print (">",i.id+"_"+args.method,file=out)
        print (seq_wrapper(dbi.query(i,method=args.method)),file=out)


if __name__=="__main__":
    from bam2x.IO import parser_factory
    p=parser_factory(description=help())
    set_parser(p)
    if len(sys.argv)==1:
        print(p.print_help())
        exit(0)
    run(p.parse_args())

 



#!/usr/bin/env python
# Programmer : zhuxp
# Date: 
# Last-modified: 02-18-2014, 14:33:50 EST
VERSION="bam2x 0.1.1"
import os,sys,argparse
from bam2x.Annotation import BED6 as Bed
from bam2x.Annotation import BED12
from bam2x import TableIO,Tools,DBI,IO
import signal
signal.signal(signal.SIGPIPE,signal.SIG_DFL)
import gzip
import time
import bam2x.Turing.TuringCodeBook as cb
from bam2x.Turing.TuringUtils import *
from bam2x.Turing import TuringCode,TuringGraph
from bitarray import bitarray
from multiprocessing import  Pool
import math
import itertools
import logging
'''
Query bam and splicing sites
Construct Turing Graph
and translate reads into BitStr

count reads number in each BitStr
v12:
penalty for increase the number of isoform
v14: 
improve: merging reads with splicing sites first. add mysort

v17: scort by (intron number + 1) * frag_numbers
v18: add rgb color

v20:
    DONE : sort by options
    DONE : bitarray to nice format rep

v20a plan:
    TODO : get coverage and add splicing sites into it. ( memory save not considered , or speed is not considered , just a rash pack)  
    
pv20b :
   parrallel computing
   DONE : splicing sites from query             (integrate xbam2splicingsites)  problem : memory control ? iterate once or twice ? 
   DONE : make it become a function 
pv20c :
    DONE : splicing sites from coverage and sequence        
    TODO : ( add Turing Coverage?)    
pv20d :
    DONE : using Turing Tuple Code instead of Turing Code
    DONE : using Bed Tuple IO ,instead of Bed IO, write Bed Tuple IO.
    DONE : simple query bam paired end (ignore the reads are not in this region)
pv20e :
    DONE: change result to json or xml
pv20f: 
    DONE : fix bugs of same location have two or more codes , revise codes length into correct one using hash [ only revise the tuple version now ] 
    DONE : strand correction
pv20g:
    DONE : adding exon structure in bam2peak?
pv20h:
    DONE : start with the exon structure.
    DONE : fix the some result entry don't have pattern error (ignore these entries)
    DONE : increase the dark of output
    DONE : ignore the pattern that is not compatible with query bed12
    DONE : add options to report seq or not
pv22a:
    DONE : report singleton as singleton , score is still cliques score
    TODO : link undetected exon.
v0.0.4:
    DONE: rm xplib and convert to bam2x lib, 
    DONE: using turing namedtuple instead of tuple
    DONE: no bam2 query but bam1 query ( fragment query in lib )
v0.1.1:
    DONE: merge only overlapping reads.  link stragdy is changed
'''
def help():
    return "detect isoforms."
def report_format(x,**kwargs):
    #TOD
    #logging.debug(x)
    try:
        S=""
        S+="QR\t"+str(x["QR"])+"\n"
        S+="FIGURE %s\n"%x["FIGURE"]
        if args.report_seq:
            if x["QR"][2]-x["QR"][1] < 100:
                S+="SQ\t"+x["WIG_TABLE"][0]
            else:
                S+="PARTIAL_SQ\t"+x["WIG_TABLE"][0][0:100]+"......."
            S+="\n"
        S+="FRG_NUMBER %s\n"%x["FRG_NUMBER"]
        S+="PATTERN_NUMBER %s\n"%x["PATTERN_NUMBER"]
        S+="  INDEX\tPATTERN\tFRGs\tINTRON_NUMBER\n"
        for i,y in enumerate(x["PATTERNS"]):
            S+="  "
            S+="No.%s\t%s\t%s\t%s\n"%(i,y[0],y[1],y[2])
        S+="CLIQUES\n"
        for i,y in enumerate(x["CLIQUES"]):
            S+="  NO."+str(i+1)+"\n"
            S+="    REP"+"\t"+y["REP"]+"\n"
            S+="    UNIQ_SCORE\t"+str(y["UNIQ_SCORE"])+"\n"
            if y.has_key("BED"):
                S+="    BED"+"\t"+str(y["BED"])+"\n"
            elif y.has_key("BEDS"):
                for bed in y["BEDS"]:
                    S+="    BED"+"\t"+str(bed)+"\n"
            if args.report_seq:
                S+="    ALL_COMPATIBLE_PATTERN\t"+str(y["ALL_GROUP"])+"\n"
                S+="    UNIQ_COMPATIBLE_PATTERN\t"+str(y["UNIQ_GROUP"])+"\n"
        S+="INTERPRET\t"+str(x["INTERPRET_FRG"])+"\n"
        S+="//\n"
        return S
    except:
        raise
        if x.has_key("QR"):
            return "# ERROR RESULT FORMAT FOR "+str(x["QR"])
        else:
            return "# UNKNOWN QR"
    


def sharpScore(l,n):
    '''
    minus nearby max score * fraction
    '''
    length=len(l)
    l2=list(l) #copy l
    #print l,n
    for i,x in enumerate(l):
        #print i,x
        near=nearby(i,n,length)
        max_nearby=0.0
        max_nearby_index=near[0]
        for j in near:
            if max_nearby < l2[j] and j!=i:
                max_nearby=l2[j]
                max_nearby_index=j
        if max_nearby > l2[i]:
            l[i]=l2[i]-float(n-abs(max_nearby_index-i))/n*max_nearby
        else:
            l[i]=l2[i]+float(n-abs(max_nearby_index-i))/n*max_nearby
        if l[i]<0: 
            l[i]=0.0
    '''
    for i,j in zip(l,l2):
        print i,j
    '''
    #print "end"
    return
    
def nearby(i,n,length):
    start=i-n
    if start<0: start=0
    end=i+n
    if end>length: end=length
    return xrange(start,end)
def nearbyIter(i,n,length,include_self=True):
    if include_self:
        yield i
    for j in range(1,n+1):
        if i-j > 0: yield i-j
        if i+j < length : yield i+j 
def set_parser(p):
    ''' This Function Parse the Argument '''
    p.add_argument('-I','--input_format',dest="format",default="bed12",choices=["bed6","bed12"],type=str,help="input file format DEFAULT:%(default)s , all possible exons")
    p.add_argument('-b','--bam',dest="bam",type=str,help="bam file")
    p.add_argument('-S','--strand',dest="strand",type=str,choices=["read1","read2"],default="read2",help="strand information")
    p.add_argument('-m','--min_uniq_percentage',dest="min_uniq",type=float,default=0.02,help="min uniq percentage [ add an isoform only if it can interpret more than min uniq percentage fragments ], DEFAULT: %(default)f ")
    p.add_argument('-f','--min_uniq_fpk_increase',dest="min_uniq_fpk_increase",type=float,default=0.05,help="default: %(default)f ")
    p.add_argument('-p','--merge_mismatch_bp',dest="merge_bp",type=int,default=5,help="default: %(default)i ")
    p.add_argument('--sort_model',dest="sort_model",type=int,default=0,choices=[0,1,2],help="model 0: sort by abundance*intron number, model 1: sort by intron number, model 2: sort by abundance, default: %(default)i ")
    p.add_argument('-g','--genome',dest="genome",type=str,help="genome sequence in 2bit format, e.g. mm9.2bit")
    p.add_argument('--report_seq',default=False,dest="report_seq",action="store_true",help="report sequence [ warning: cost mem ]")
def run(local_args):
    logging.basicConfig(level=logging.WARNING)
    global args,out,dbi_bam,g, MIN_INTRON_LENGTH, MIN_SPLICING_SITES_SCORE, MIN_FPK_RATIO,query_num
    MIN_INTRON_LENGTH=10
    MIN_SPLICING_SITES_SCORE=2
    '''
    IO TEMPLATE
    '''
    '''
    mySorts={ 0:sort_by_intron_and_abundance,
              1:sort_by_intron,
              2:sort_by_abundance
    }
    '''
    args=local_args
    #print "debug:",args.report_seq
    MIN_FPK_RATIO=args.min_uniq_fpk_increase #TO TEST
    fin=IO.fopen(args.input,"r")
    out=IO.fopen(args.output,"w")
    '''
    END OF IO TEMPLATE 
    '''
    print >>out,"# This data was generated by program ",sys.argv[0]," (version: %s)"%VERSION,
    print >>out,"in bam2x ( https://github.com/nimezhu/bam2x )"
    print >>out,"# Date: ",time.asctime()
    print >>out,"# The command line is :"
    print >>out,"#\t"," ".join(sys.argv)
    # header=["chr","start","end","id","score","strand","seq"];
    # dbi_splicing_sites=DBI.init(args.splicing_sites,"tabix",tabix="metabed",header=header);
    #if args.format=="guess":
    #    args.format=IO.guess_format(args.input)
    reader=TableIO.parse(fin,args.format)
    query_list=[]
    query_lists=[[] for i in range(args.num_cpus)]
    query_num=0
    for i,x in enumerate(reader):
        query_lists[i%args.num_cpus].append(x)
    query_num=i+1
    #querys(query_lists[0])  #DEBUG
    pool=Pool(processes=args.num_cpus)
    results=pool.map(querys,query_lists)
    #print results
    output(results)


def querys(listx):
    dbi_bam=DBI.init(args.bam,"bam")
    genome=DBI.init(args.genome,"genome")
    results=[]
    for x in listx:
        results.append(query(x,dbi_bam,genome))
    return results

def output(results):
    for i in iter(results):
        print >>out,report_format(i,indent="  ")
def iter(results):
    index=-1
    #print "in iter"
    #print "query_num:",query_num
    for i in range(query_num):
        j=i%args.num_cpus
        if j==0:
            index+=1
        #print"in iter:",results[j][index]
        yield results[j][index]


def query(i,dbi_bam,genome): # i is query iteem
    h={}
    hc={}
    #retv="" #RETURN VALUE
    #retv+="BEGIN\nQR\t"+str(i)+"\n"
    x=i
    ret_dict={}
    ret_dict["QR"]=i
    '''
    PART A: scoring and adding splicing sites
    '''
    print >>sys.stderr,"QR",str(i)
    length=i.stop-i.start
    g=list()
    g.append(TuringCode(0,cb.ON))
    g.append(TuringCode(0,cb.BLOCKON))
    g.append(TuringCode(length,cb.OFF))
    g.append(TuringCode(length,cb.BLOCKOFF))
    #logging.debug(g)
    array=[0.0 for j in xrange(length)]
    donorSitesScore=[0.0 for j in xrange(length+1)]
    acceptorSitesScore=[0.0 for j in xrange(length+1)]
    seq=genome.get_seq(i).upper()
    for j in dbi_bam.query(i,method="bam1",strand=args.strand):
        read=Tools.translate_coordinates(i,j)
        #TODO : get read anc call peaks
        if read.strand=="+":
            for k in read.Exons():
                for k0 in xrange(k.start,k.stop):
                    if k0>=length:
                        break
                    if k0>=0:
                        array[k0]+=1
	    for intron in read.Introns():
	        if len(intron)< MIN_INTRON_LENGTH: continue
                if intron.start > 0  and intron.start <length:
                    donorSitesScore[intron.start]+=1
                if intron.stop > 0  and intron.stop <length:
                    acceptorSitesScore[intron.stop]+=1
    
    gt=[0 for j in range(length)]
    ag=[0 for j in range(length)]
    for j in xrange(length-1):
        if seq[j]=="G" and seq[j+1]=="T":
            gt[j]=1
        elif seq[j]=="A" and seq[j+1]=="G":
            ag[j]=1
    '''
    Score SplicingScores
    '''
    #DONE CLean Scores [ Minus the near by scores]
    sharpScore(donorSitesScore,args.merge_bp)
    sharpScore(acceptorSitesScore,args.merge_bp)
   
    hDonor={}
    hAcceptor={}
    hAcceptor[0]=1
    hDonor[length]=1
    for j in xrange(length):
        if donorSitesScore[j] > MIN_SPLICING_SITES_SCORE:
            g.append(TuringCode(j,cb.BLOCKOFF))
            hDonor[j]=1
        elif acceptorSitesScore[j] > MIN_SPLICING_SITES_SCORE:
            g.append(TuringCode(j,cb.BLOCKON))
            hAcceptor[j]=1
    '''
    closure to score sites by coverage
    '''
    
    def addCoverageAndSeqScore():
        '''
        two step, 
        1. phase change threshold determin
        2. find the nearest possible splicing sites 
        '''
        #logging.debug(length)
        #logging.debug(g)
        diff=[array[0]] # assume the pos -1  is 0.0
        for j in range(length-1):
            diff.append(math.log((array[j+1]+1)/(array[j]+1.0)))
        def mean_sd(indexes):
            n=0
            s=0.0
            sd=0.0
            a=[]
            for k in indexes:
                n+=1
                max_diff=0.0
                for k0 in nearbyIter(k,args.merge_bp,length):
                    if k0>=length: continue
                    if abs(max_diff) < abs(diff[k0]):
                        max_diff=diff[k0]
                s+=max_diff
                a.append(max_diff)
            if n>0: 
                mean=s/n
            else:
                return 0.0,0.0
            for k in a:
                sd+=(k-mean)*(k-mean)
            if n > 1:
                sd=sd/(n-1)
            if n > 1:
                sd=math.sqrt(sd/(n-1))
            else:
                sd=0.0
            return mean,sd

        (meanDonor,sdDonor)=mean_sd(hDonor.keys())
        (meanAcceptor,sdAcceptor)=mean_sd(hAcceptor.keys())
        #print meanDonor,sdDonor
        #print meanAcceptor,sdAcceptor
        '''
        adding possible splicing sites based on diff value
        '''
        def correctToNearAcceptor(j):
            signal=0
            for k in nearbyIter(j,args.merge_bp,length):
                if hAcceptor.has_key(k): 
                    return k
            for k in nearbyIter(j,args.merge_bp,length):
                try:
                    if ag[k]==1:
                        return k+2
                except:
                    pass
            return -1
        def correctToNearDonor(j):
            signal=0
            for k in nearbyIter(j,args.merge_bp,length):
                if hDonor.has_key(k): 
                    return k
            for k in nearbyIter(j,args.merge_bp,length):
                try:
                    if gt[k]==1:
                        return k
                except:
                    pass
            return -1


        for j in range(length):
            #if diff[j]  > sdAcceptor + meanAcceptor:
            if diff[j]  > meanAcceptor + sdAcceptor and diff[j] > 0.0:
                k=correctToNearAcceptor(j)
                if k!=-1 and not hAcceptor.has_key(k):
                    hAcceptor[k]=1
                    g.append(TuringCode(k,cb.BLOCKON))
                # add to revise small fragrments ?
                '''
                elif k==-1 and diff[j] > meanAcceptor + 3 * sdAcceptor and not hAcceptor.has_key(j):
                    hAcceptor[j]=1
                    l.append((j,cb.BLOCKON))
                '''


            #if  diff[j] < meanDonor - sdDonor:
            if  diff[j] < meanDonor -  sdDonor and diff[j] < 0.0:
                k=correctToNearDonor(j)
                if k!=-1 and not hDonor.has_key(k):
                    hDonor[k]=1
                    g.append(TuringCode(k,cb.BLOCKOFF))
                '''
                elif k==-1 and diff[j] <  meanDonor - 3 * sdDonor and not hDonor.has_key(j):
                    hDonor[j]=1
                    l.append((j,cb.BLOCKOFF))
                '''
        if isinstance(x,BED12): #strange could not use i?
            ti=Tools.translate_coordinates(x,x)
            #print "debug ti:",tia
            ti_codes=[]
            for j in ti.blockStarts:
                k=correctToNearAcceptor(j)
                if k==-1:
                    ti_codes.append(TuringCode(j,cb.BLOCKON))
                else:
                    ti_codes.append(TuringCode(k,cb.BLOCKON))
                if k!=-1 and not hAcceptor.has_key(k):
                    hAcceptor[k]=1
                    g.append(TuringCode(k,cb.BLOCKON))
                if k==-1 and not hAcceptor.has_key(j):
                    hAcceptor[j]=1
                    g.append(TuringCode(j,cb.BLOCKON))
            last_stop=0
            for j,m in itertools.izip(ti.blockStarts,ti.blockSizes):
                blockStop=j+m
                #print "debug blockStop",blockStop
                k=correctToNearDonor(blockStop)
                if k==-1:
                    ti_codes.append(TuringCode(blockStop,cb.BLOCKOFF))
                    last_stop=j
                else:
                    ti_codes.append(TuringCode(k,cb.BLOCKOFF))
                    last_stop=k
                if k!=-1 and not hDonor.has_key(k):
                    hDonor[k]=1
                    g.append(TuringCode(k,cb.BLOCKOFF))
                elif k==-1 and not hDonor.has_key(blockStop):
                    hDonor[blockStop]=1
                    g.append(TuringCode(blockStop,cb.BLOCKOFF))

            ti_codes.append(TuringCode(correctToNearAcceptor(ti.blockStarts[0]),cb.ON))
            ti_codes.append(TuringCode(last_stop,cb.OFF))
            ti_codes.sort()
            #print "debug ti_codes",ti_codes
            #print ti_codes
            #TODO  GENERate start path.
            # l.sort()
            graph=TuringGraph(g)
            g_len=len(graph)
            initial_bits=graph.translate_path_into_bits(TuringGraph(ti_codes),args.merge_bp)
            #print "debug ", initial_bits
            for i in range(0,2*g_len,2):
                if initial_bits[i]==True and initial_bits[i+1]==False:
                    initial_bits[i+1]=True
            initial_bits[-1]=True
            initial_bits[-2]=True
            return initial_bits,graph
        # l.sort()
        return None,TuringGraph(g)




            


    
    initial_bits,g=addCoverageAndSeqScore()
    '''
    PART B : Report wig
    retv+="WIG"+"\n"
    retv+= "index\tnt\tcoverage\tdonorSitesScore\tacceptorSitesScore\tGT\tAG\n"
    for j in xrange(len(i)):
        retv+= str(j)+"\t"+str(seq[j])+"\t"+str(array[j])+"\t"+str(donorSitesScore[j])+"\t"+str(acceptorSitesScore[j])+"\t"+str(gt[j])+"\t"+str(ag[j])+"\n"
    '''
    #logging.debug(initial_bits)
    if args.report_seq:
        ret_dict["WIG_TABLE"]=(seq,array,donorSitesScore,acceptorSitesScore,gt,ag)
    
    '''
    end of adding wig
    '''
    
    
    
    '''
    PART C: GREEDY ADDING ISOFORMS
    
    '''
    
    
    # bitarray_path=bitarray(2*len(g))
    # bitarray_path.setall(True)
    #TODO change bitarray?
    #logging.debug(g)
    ret_dict["FIGURE"]=g.graph_str(600)
    h={}
    hc={}
    j0=0;
    total_frag=0
    #for j in g: print "debug g:",j
    #print "INIT",bitarray_to_rep(initial_bits)
    g_len=len(g)
    if initial_bits is None:
        initial_bits=bitarray(2*g_len)
        initial_bits.setall(True)

    for j in dbi_bam.query(i,method="bam2fast",strand=args.strand):
        p=[]
        #logging.debug(j)
        #print "debug",j
        
        if j[0].strand!=i.strand: continue
        for k in j:
            p.append(TuringFactory(Tools.translate_coordinates(i,k))) #TODO check this
        # p.append(TuringFactory(Tools.translate_coordinates(i,j))) #TODO check this

        #print "debug glen:",g_len

        a=g.translate_paths_into_bits(p,args.merge_bp)
        #print "debug bits: ",a
        #print "debug bed :",translate_bits_into_bed(g,a)
        if isSelfIncompatible(a): continue
        if h.has_key(a.tobytes()):
            h[a.tobytes()]+=1
            if hc[a.tobytes()]!=a:
                print >>sys.stderr,"WARNING"
        else:
            h[a.tobytes()]=1
            hc[a.tobytes()]=a
        j0+=1
     #retv+="PATTERN_NUMBER\t"+str(len(h.keys()))+"\n"
    ret_dict["PATTERN_NUMBER"]=len(h.keys())
    # print >>sys.stderr,"debug pattern number",ret_dict["PATTERN_NUMBER"]
    #retv+="FRG_NUMBER\t"+str(j0)+"\n"
    ret_dict["FRG_NUMBER"]=j0
    total_frag=j0
    if j0==0: return ret_dict
    sorted_keys=h.keys()
    if args.sort_model==0:
        sorted_keys.sort(lambda x,y:bitarray_to_intron_number(hc[x])*h[x]-bitarray_to_intron_number(hc[y])*h[y]  or h[x]-h[y], reverse=True)
    elif args.sort_model==1:
        sorted_keys.sort(lambda x,y:bitarray_to_intron_number(hc[x])-bitarray_to_intron_number(hc[y]) or h[x]-h[y], reverse=True)
    else:
        sorted_keys.sort(lambda x,y:h[x]-h[y] or bitarray_to_intron_number(hc[x])-bitarray_to_intron_number(hc[y]), reverse=True) 
    clique=[]
    cliques=[]
    bits_no_prior=bitarray(g_len*2)
    bits_no_prior.setall(True)
    cliques_pattern=[]
    ret_dict["PATTERNS"]=list()

    keys_len=len(sorted_keys)
    
    
    joined_cliques=[False for i0 in range(keys_len)]
    unjoined_pattern=range(0,keys_len)

    for j,key in enumerate(sorted_keys):
        ret_dict["PATTERNS"].append((bitarray_to_rep(hc[key]),h[key],bitarray_to_intron_number(hc[key])))
        if joined_cliques[j] : continue
        joined_clique=False
        joined_cliques[j]=True
        bits_no_prior=bitarray(g_len*2)
        bits_no_prior.setall(True)  
        h_clique={}
        if isCompatible(initial_bits,hc[sorted_keys[j]]):
            max_index=0
            clique=[]
            clique.append(j)
            cliques.append(clique)
            cliques_pattern.append(hc[sorted_keys[j]])
            h_clique[j]=1
            update=True
            while update:
                update=False
                logging.debug("update:%s",update)
                for k0,k in enumerate(unjoined_pattern):
                    logging.debug("del %i",k)
                    if k==j:
                        del unjoined_pattern[k0]
                    elif isOverlapCompatible(cliques_pattern[-1],hc[sorted_keys[k]]):
                        update=True
                        cliques[-1].append(k)
                        cliques_pattern[-1]=bitarray_and(cliques_pattern[-1],hc[sorted_keys[k]])
                        joined_cliques[k]=True
                        del unjoined_pattern[k0]
                        h_clique[k]=1
                        break
                if update==False:
                    if k==j:
                        continue
                    for k in range(keys_len):
                        if h_clique.has_key(k): continue
                        if isOverlapCompatible(cliques_pattern[-1],hc[sorted_keys[k]]):
                            update=True
                            cliques[-1].append(k)
                            cliques_pattern[-1]=bitarray_and(cliques_pattern[-1],hc[sorted_keys[k]])
                            h_clique[k]=1
                            break



    ret_dict["CLIQUES"]=list()

    cumulative_score=0
    #update all cliques pattern.
    j0=0
    #buffer_keys=[] # those key are ignored in previous uniqs
    max_uniq_fpk=0.0
    max_uniq=0
    for j,x in enumerate(cliques):
        score=0
        
        cliques_pattern[j][-1]=True
        cliques_pattern[j][-2]=True

        uniq_score=0
        for k,y in enumerate(x):
            uniq_score+=h[sorted_keys[y]]
        '''
        for k,y in enumerate(buffer_keys):
            if isCompatible(cliques_pattern[buffer_keys[k]],
                cliques[j].append(k)
        '''
        if float(uniq_score)/total_frag < args.min_uniq :
            #print >>sys.stderr,"ignore due to small uniq frags:",float(uniq_score)/total_frag
            continue
        else:
            logging.debug(cliques_pattern)
            bed=g.translate_bits_into_bed(cliques_pattern[j])
            cdna_length=bed.cdna_length()
            if cdna_length==0:
                #print >>sys.stderr,"ignore due to cdna_length",bed
                continue
            uniq_fpk=float(uniq_score)/cdna_length*1000.0
            if  uniq_fpk > max_uniq_fpk:
                max_uniq_fpk=uniq_fpk
            if uniq_fpk  < max_uniq_fpk * MIN_FPK_RATIO:
                #print >>sys.stderr,"ignore due to fpk",bed,"\tuniq_fpk:",uniq_fpk,"\t current max:",max_uniq_fpk
                continue
            rgb=200-uniq_score*200/total_frag  
            cumulative_score+=uniq_score
            j0+=1
            #retv+="NO."+str(j0)+" CLIQUE"+"\n"    
            #retv+="CLIQUE\t"+str(c)+"\nUNIQ\t"+str(cliques[j])+"\nPATTERN\t"+bitarray_to_rep(cliques_pattern[j])+"\n"
            ret_dict["CLIQUES"].append(dict())
            #ret_dict["CLIQUES"][-1]["CLIQUE"]=(c,cliques[j],bitarray_to_rep(cliques_pattern[j]))
            ret_dict["CLIQUES"][-1]["ALL_GROUP"]=cliques[j]
            ret_dict["CLIQUES"][-1]["UNIQ_GROUP"]=cliques[j]
            ret_dict["CLIQUES"][-1]["REP"]=bitarray_to_rep(cliques_pattern[j])
            #retv+="CLIQUE\t",c,"\nUNIQ\t",cliques[j],"\nPATTERN\t",cliques_pattern[j]
            '''
            pattern
            need to revise
            '''
            pattern=cliques_pattern[j]
            pattern[-1]=True
            pattern[-2]=True
            beds=[bed for bed in g.translate_bits_into_beds(cliques_pattern[j])]
            ret_dict["CLIQUES"][-1]["BEDS"]=[]
            for j1,bed in enumerate(beds):
                #logging.debug(i)
                score=score*1000.0/cdna_length
                chr=i.id
                id=i.id+"_"+"NO."+str(j0)+"_"+str(j1)
                itemRgb=str(rgb)+","+str(rgb)+","+str(rgb)
                bed=bed._replace(chr=chr,id=id,itemRgb=itemRgb)
                ret_dict["CLIQUES"][-1]["BEDS"].append(str(Tools.translate_coordinates(i,bed,True)))

           


            ret_dict["CLIQUES"][-1]["UNIQ_SCORE"]=uniq_score
            ret_dict["CLIQUES"][-1]["UNIQ_FPK"]=uniq_fpk
            ret_dict["CLIQUES"][-1]["SCORE"]=score

            #retv+="UNIQ\t"+str(uniq_score)+"\tBED\t"+str(Tools.translate_coordinates(Bed(i),bed,True))+"\n"
            #print >>sys.stderr,"UNIQ_FPK/CURRENT_MAX_UNIQ_FPK",uniq_fpk/max_uniq_fpk

            #retv+="\n"
    ret_dict["INTERPRET_FRG"]=float(cumulative_score)/total_frag
    return ret_dict

    



    


if __name__=="__main__":
    run()




from __future__ import print_function
import os
import sys
import logging
import argparse
import pysam
from bam2x import IO,TableIO
from os.path import splitext
from string import digits
from pysam import tabix_index
def help():
    return "tabix index a bed or vcf file"
def set_parser(parser):
    parser.add_argument("-S","--sorted",dest="sorted",action="store_true")
    parser.add_argument("-I","--format",dest="format",type=str,choices={"bed12","bed6","bed3","vcf"},default="bed12")
    #parser.add_argument("-s","--seq",type=int,dest="seq",default=None)
    #parser.add_argument("-b","--begin",type=int,dest="begin",default=None)
    #parser.add_argument("-e","--end",type=int,dest="end",default=None)
    #parser.add_argument("-Z","--zero_index",action="store_true",dest="zero",default=True)
    
    
def run(args):
    fin=IO.open(args.input,"r")
    outfile=args.input
    if not args.sorted:
        l = [ i for i in TableIO.parse(fin,args.format) ]
        l.sort()
        name=splitext(args.input)
        outfile = "{name[0]}.sorted{name[1]}".format(name=name)
        out = IO.open(outfile,"w")
        for i in l:
            print(i,file=out)
        out.close()
    format=args.format.translate(None,digits)
    tabix_index(outfile,preset=format)
    

if __name__=="__main__":
    from bam2x.IO import parser_factory
    p=parser_factory(description=help())
    set_parser(p)
    if len(sys.argv)==1:
        print(p.print_help())
        exit(0)
    run(p.parse_args())










#!/usr/bin/env python
# Programmer : zhuxp
# Date: 
# Last-modified: 02-13-2014, 14:53:10 EST
VERSION="0.0.9"
import os,sys,argparse
from bam2x.Annotation import BED6 as Bed
from bam2x.Annotation import BED3
from bam2x.Annotation import BED12 as Bed12
from bam2x import TableIO,Tools,DBI,IO
import signal
signal.signal(signal.SIGPIPE,signal.SIG_DFL)
import gzip
import time
import multiprocessing as mp
import array,tempfile,heapq
import bam2x.prob as prob
import itertools
assert array.array('i').itemsize==4

MIN_GAP_IN_SAME_EXON=10
'''
V3:
DONE: mv sorting array to Turing Module
V4:
DONE: correct the coverage calculation
DONE: report link exon and intron

V5:
DONE: same memory using numarray instead of bed?
      only using array?
      rm bed class?
      list of list ?
      bedgraph sorting array?
DONE: report if peak has intron and the possible cDNA length and gene length.
V6:
    change format to Bed12
    get rid of terminal introns
V7:
    don't change state because of intron just link all exon will do.
    get rid of consective small exons.
V8:
    DONE: add complexity filter out [ LZW complexity / length > 0.4 ]
    DONE: mv to bam2x subcommand
V9:
    DONE: rm xplib?
    DONE: mv to namedtuple turing

TODO: compare with known gene
TODO: trim the last intron or extend the exon? ( KEY PROBLEM. how to define the end )
'''
EXON_GROUP_CODE=1
INTRON_GROUP_CODE=0
HAS_INTRON=20
NOT_HAS_INTRON=10

POSITIVE_STRAND=1
NEGATIVE_STRAND=-1

START_INDEX=0
STOP_INDEX=1
SCORE_INDEX=2
STRAND_INDEX=3
GROUP_INDEX=4
OTHER_INDEX=5
ID_INDEX=6
EXONSTARTS_INDEX=7
EXONSIZES_INDEX=8
CPLX_CUTOFF=0.4
def help():
    return "bam2peak , call exon peaks and link exons"
def set_parser(p):
    ''' This Function Parse the Argument '''
    p.add_argument('-s','--strand',dest="strand",type=str,choices=["read1","read2"],default="read2",help="read1 or read2 is the positive strand , default: %(default)s")
    p.add_argument('--prefix',dest="prefix",type=str,default="R",help="prefix for bed name default: %(default)s")
    p.add_argument('--cutoff',dest="cutoff",type=int,default=1,help="only report region covrage >= cutoff : %(default)i")
    p.add_argument('--pvalue',dest="pvalue",type=float,default=1e-05,help="cutoff for calling peak : %(default)f")
    p.add_argument('-g','--genome',dest="genome",type=str,help="genome sequence file in twobit format")
def run(args_local):
    '''
    IO TEMPLATE
    '''
    global args,out,exon_cutoff,intron_cutoff,hasGenome
    args=args_local
    dbi=DBI.init(args.input,"bam")
    out=IO.fopen(args.output,"w")
    hasGenome=False
    if args.genome:
        hasGenome=True
    
    '''
    END OF IO TEMPLATE 
    '''
    print >>out,"# This positive_data was generated by program ",sys.argv[1]," (version: %s)"%VERSION,
    print >>out,"in bam2x ( https://github.com/nimezhu/bam2x )"
    print >>out,"# Date: ",time.asctime()
    print >>out,"# The command line is :"
    print >>out,"#\t"," ".join(sys.argv)
    chrs=[]
    lengths=[]
    for i in dbi.bamfiles[0].references:
        chrs.append(i)
    for i in dbi.bamfiles[0].lengths:
        lengths.append(i)
    p=mp.Pool(processes=args.num_cpus)    
    coverage_bedgraphs=p.map(process_chrom,chrs)
    bedgraphs=[]
    coverages=[]
    for i in range(len(chrs)):
        bedgraphs.append(coverage_bedgraphs[i][1])
        coverages.append(coverage_bedgraphs[i][0])
    s=0.0  # 1000.0
    l=long(0)
    for i in range(len(chrs)):
        s+=coverages[i]
        l+=lengths[i]
    l=l*2 # Double Strand
    coverage=s/l*1000.0
    threshold=1
    while 1:
        if prob.poisson_cdf(threshold,coverage,False) < args.pvalue: break
        threshold+=1
    exon_cutoff=threshold
    intron_cutoff=2 #TODO revise it
    print >>out,"# MEAN COVERAGE:",coverage
    print >>out,"# EXON COVERAGE CUTOFF:",exon_cutoff
    
    #call_peaks(bedgraphs[0],1) #debug
    peaks=p.map(call_peaks_star,itertools.izip(chrs,bedgraphs,itertools.repeat(exon_cutoff)))
    output(chrs,peaks)

    #process_chrom("chr1")
def output(chrs,s):
    for chrom,i in itertools.izip(chrs,s):
        for j in i:
            print >>out,nice_format(chrom,j)

def nice_format(chrom,a):
    s=chrom+"\t"
    s+=str(a[START_INDEX])+"\t"
    s+=str(a[STOP_INDEX])+"\t"
    s+=chrom+"_"+str(a[ID_INDEX])+"\t"
    s+=str(a[SCORE_INDEX])+"\t"
    if a[STRAND_INDEX]==1:
        s+="+\t"
    elif a[STRAND_INDEX]==-1:
        s+="-\t"
    else:
        s+=".\t"
    s+=str(a[START_INDEX])+"\t" #cds_start
    s+=str(a[START_INDEX])+"\t" #cds_stop
    s+="0,0,0\t" #Rgb
    s+=str(len(a[EXONSIZES_INDEX]))+"\t"
    for i in a[EXONSIZES_INDEX]:
        s+=str(i)+","
    s+="\t"
    for i in a[EXONSTARTS_INDEX]: 
        s+=str(i)+","
    
    return s

    #return "\t".join("%s"%item for item in a)
from bam2x.Turing import TuringCode
from bam2x.Turing import TuringCodeBook as cb
from bam2x.Turing import TuringSortingArray
from operator import itemgetter
from bam2x.Tools.LZW import complexity 
def process_chrom(chrom):
    local_dbi=DBI.init(args.input,"bam")
    retv=list()
    intron_retv=list()
    a=[]
    #positive_data=TuringSortingArray(None,500)
    positive_data=TuringSortingArray()
    negative_data=TuringSortingArray()
    
    positive_intron_data=TuringSortingArray()
    negative_intron_data=TuringSortingArray()

    for i in local_dbi.query(chrom,method="bam1",strand=args.strand):
        if i.strand=="+" or i.strand==".": 
            positive_data.append(TuringCode(i.start,cb.ON))
            positive_data.append(TuringCode(i.stop,cb.OFF))
            for j in i.Exons():
                positive_data.append(TuringCode(j.start,cb.BLOCKON))
                positive_data.append(TuringCode(j.stop,cb.BLOCKOFF))
            for j in i.Introns():
                positive_intron_data.append(TuringCode(j.start,cb.BLOCKON))
                positive_intron_data.append(TuringCode(j.stop,cb.BLOCKOFF))
        else:
            negative_data.append(TuringCode(i.start,cb.ON))
            negative_data.append(TuringCode(i.stop,cb.OFF))
            for j in i.Exons():
                negative_data.append(TuringCode(j.start,cb.BLOCKON))
                negative_data.append(TuringCode(j.stop,cb.BLOCKOFF))
            for j in i.Introns():
                negative_intron_data.append(TuringCode(j.start,cb.BLOCKON))
                negative_intron_data.append(TuringCode(j.stop,cb.BLOCKOFF))
    cutoff=args.cutoff
    coverage=0.0
    for i,x in enumerate(codesToBedGraph(positive_data.iter())):
        if x[SCORE_INDEX] >= cutoff:
            retv.append((x[0],x[1],x[2],POSITIVE_STRAND,EXON_GROUP_CODE))
        coverage+=float(x[1]-x[0])*x[2]/1000.0
    for i,x in enumerate(codesToBedGraph(negative_data.iter())):
        if x[SCORE_INDEX] >= cutoff:
            retv.append((x[0],x[1],x[2],NEGATIVE_STRAND,EXON_GROUP_CODE))
        coverage+=float(x[1]-x[0])*x[2]/1000.0
    INTRON_CUTOFF=1.0
    for i,x in enumerate(codesToBedGraph(positive_intron_data.iter())):
        if x[SCORE_INDEX] >= INTRON_CUTOFF:
            retv.append((x[0],x[1],x[2],POSITIVE_STRAND,INTRON_GROUP_CODE))
    for i,x in enumerate(codesToBedGraph(negative_intron_data.iter())):
        if x[SCORE_INDEX] >= INTRON_CUTOFF:
            retv.append((x[0],x[1],x[2],NEGATIVE_STRAND,INTRON_GROUP_CODE))
    retv.sort(key=itemgetter(0,1,2))
    #TODO how to sort!
    local_dbi.close()
    return coverage,retv
def codesToBedGraph(iter):
    a=iter.next()
    last_pos=a[0]
    counter=0
    for i in iter:
        if i.pos!=last_pos:
            #TODO using more efficient data structure
            yield (last_pos,i.pos,counter)
            last_pos=i.pos
        if i.code==cb.BLOCKON:
            counter+=1
        if i.code==cb.BLOCKOFF:
            counter-=1
    raise StopIteration


def call_peaks_star(a_b_c):
    return call_peaks(*a_b_c)
    
def call_peaks(chrom,bedgraph,exon_cutoff):
    #TODO
    
    if hasGenome:
        genome=DBI.init(args.genome,"genome")
    def filter_low_complexity(peak):
        new_exon_starts=[]
        new_exon_sizes=[]
        local_peak=list(peak)
        offset=local_peak[START_INDEX]
        for exon_start,exon_size in itertools.izip(local_peak[EXONSTARTS_INDEX],local_peak[EXONSIZES_INDEX]):
            exon_s=offset+exon_start
            exon_e=offset+exon_start+exon_size
            seq=genome.query(BED3(chrom,exon_s,exon_e),method="seq")
            if len(seq) > 100:
                cplx=float(complexity(seq[0:100]))/100
            else:
                cplx=float(complexity(seq))/len(seq)
            if cplx > CPLX_CUTOFF:
                new_exon_starts.append(exon_start)
                new_exon_sizes.append(exon_size)
        if len(new_exon_sizes) > 0:
            local_peak[EXONSIZES_INDEX]=tuple(new_exon_sizes)
            local_peak[START_INDEX]=new_exon_starts[0]+offset
            local_peak[STOP_INDEX]=new_exon_starts[-1]+offset+new_exon_sizes[-1]
            shift_start=new_exon_starts[0]
            for i0 in range(len(new_exon_starts)):
                new_exon_starts[i0]-=shift_start
            local_peak[EXONSTARTS_INDEX]=tuple(new_exon_starts)
            return tuple(local_peak)
        else:
            return None
    gap=10
    pos_beds=[]
    neg_beds=[]
    peaks=[]
    i_p=0
    i_n=0
    last_pos_stop=0
    last_neg_stop=0
    for i in bedgraph:
        if i[STRAND_INDEX]==POSITIVE_STRAND:
            if i[GROUP_INDEX]==EXON_GROUP_CODE:
                if i[SCORE_INDEX] >= exon_cutoff:
                    if len(pos_beds)>0:
                        if  i[START_INDEX]-last_pos_stop < gap or last_pos_stop==0:
                            pos_beds.append(i)
                            if last_pos_stop < i[STOP_INDEX]:
                                last_pos_stop=i[STOP_INDEX]
                        else:
                            
                            peak=bedsToPeak(pos_beds,"p_"+str(i_p))
                            if peak is not None:
                                if hasGenome:
                                    peak=filter_low_complexity(peak)
                                    if peak is not None:
                                        peaks.append(peak)
                                        i_p+=1
                                else:
                                    peaks.append(peak)
                                    i_p+=1
                            pos_beds=[i]
                            last_pos_stop=i[STOP_INDEX]
                    else:
                        last_pos_stop=i[STOP_INDEX]
                        pos_beds.append(i)
            elif i[GROUP_INDEX]==INTRON_GROUP_CODE:
                if last_pos_stop < i[STOP_INDEX]:
                    last_pos_stop=i[STOP_INDEX]
                pos_beds.append(i)
        else:
            if i[GROUP_INDEX]==EXON_GROUP_CODE:
                if i[SCORE_INDEX] >= exon_cutoff:
                    if len(neg_beds)>0:
                        if  i[START_INDEX]-last_neg_stop < gap or last_neg_stop==0:
                            neg_beds.append(i)
                            if last_neg_stop < i[STOP_INDEX]:
                                last_neg_stop=i[STOP_INDEX]
                        else:
                            peak=bedsToPeak(neg_beds,"n_"+str(i_n))
                            if peak is not None:
                                if hasGenome:
                                    peak=filter_low_complexity(peak)
                                    if peak is not None:
                                        peaks.append(peak)
                                        i_n+=1
                                else:
                                    peaks.append(peak)
                                    i_n+=1
                            neg_beds=[i]
                            last_neg_stop=i[STOP_INDEX]
                    else:
                        neg_beds.append(i)
                        last_neg_stop=i[STOP_INDEX]
            elif i[GROUP_INDEX]==INTRON_GROUP_CODE:
                if last_neg_stop < i[STOP_INDEX]:
                    last_neg_stop=i[STOP_INDEX]
                neg_beds.append(i)

    if len(pos_beds)>0:
        peak=bedsToPeak(pos_beds,"p_"+str(i_p))
        if peak is not None:
            if hasGenome:
                peak=filter_low_complexity(peak)
                if peak is not None:        
                    peaks.append(peak)
            else:
                peaks.append(peak)
    if len(neg_beds)>0:
        peak=bedsToPeak(neg_beds,"n_"+str(i_n))
        if peak is not None:
            if hasGenome:
                peak=filter_low_complexity(peak)
                if peak is not None:        
                    peaks.append(peak)
            
            else:
                peaks.append(peak)
    peaks.sort()
    return peaks
def length(x):
    return x[STOP_INDEX]-x[START_INDEX]
def bedsToPeak(ibeds,id):
    try:
        while(ibeds[0][GROUP_INDEX]==INTRON_GROUP_CODE):
            try:
                ibeds=ibeds[1:]
            except:
                return None
        peak=[ibeds[0][START_INDEX],ibeds[0][STOP_INDEX],float(ibeds[0][SCORE_INDEX]),ibeds[0][STRAND_INDEX],NOT_HAS_INTRON,0,id]
        cdna_length=length(peak)
        exonstarts=[]
        exonsizes=[]
        exon_start_point=ibeds[0][START_INDEX]
        exon_stop_point=ibeds[0][STOP_INDEX]
        exon_signal=1
        for i in ibeds[1:]:
            if i[GROUP_INDEX]==EXON_GROUP_CODE:
                if exon_signal==0:
                    exon_start_point=i[START_INDEX]
                    exon_signal=1
                elif i[START_INDEX]-exon_stop_point > MIN_GAP_IN_SAME_EXON:
                    exonstarts.append(exon_start_point-peak[0])
                    exonsizes.append(exon_stop_point-exon_start_point)
                    #print "debug in intron"
                    exon_start_point=i[START_INDEX]
                    exon_stop_point=i[STOP_INDEX]

                peak[SCORE_INDEX]=float(peak[SCORE_INDEX]*cdna_length+i[SCORE_INDEX]*length(i))/(length(peak)+length(i))
                peak[STOP_INDEX]=i[STOP_INDEX]
                cdna_length+=length(i)
                exon_stop_point=i[STOP_INDEX]
            else:
                peak[GROUP_INDEX]=HAS_INTRON
                '''
                if exon_signal==1:
                    exonstarts.append(exon_start_point-peak[0])
                    exonsizes.append(exon_stop_point-exon_start_point)
                    #print "debug in intron"
                    exon_signal=0
                '''
        if exon_signal==1:
            exonstarts.append(exon_start_point-peak[0])
            exonsizes.append(exon_stop_point-exon_start_point)
            exon_signal=0
        peak[OTHER_INDEX]=cdna_length
        '''
        print "DEBUG META",peak
        for i in ibeds:
            print "DEBUG IN",i
        print "DEBUG",exonstarts
        print "DEBUG",exonsizes
        '''
        peak.append(tuple(exonstarts))
        peak.append(tuple(exonsizes))
        #print "debug",tuple(peak)
        return tuple(peak)
    except:
        # ibeds is zero.
        return None



        

    
if __name__=="__main__":
    run()








